/*** 
 * 		General Information
 * 
 * Target-MCU: any Atmel 8-bit MCU providing:  16-bit Timer with Compare Match Unit, Hardware USART with 250kbps
 * Hardware Usage: one USART, one 16bit-Timer, about 1KB of program-flash
 * Needs: Interrupts enabled, enough processing time ;)
 * Gives: DMX RX and/or TX with only one USART at the same time
 * HowToUse RX:
 * 		+ set CHANNEL_TO_RX to number of channels requested to receive. Set to 0 to disable the receiver
 * 		+ Set a startaddress to DmxRx.startAddr
 * 		+ start with dmx_init()
 * 		+ Get received data out of DmxRx.values[x], where x is the requested channel after startaddress (beginning with 0)
 * 		+ Get status (DMX comming or not) out of DmxRx.comming. There is data if DmxRx.comming > 0
 * HowToUse TX:
 * 		+ set CHANNEL_TO_TX to number of channels requested to send. Set to 0 to disable the transmitter
 * 		+ start with dmx_init()
 * 		+ Write data to send into DmxTx.values[x], where x is the DMX-channel number (counting from 0)
 * 
 * License: Creative Commons 3.0 BY-NC-SA
 * Author: Jan Wiese <jw-lighting@ewetel.net>
 * Date: 30.07.2011
 * Version: 1.1 stable
 * 
 *  This software comes with absolutly NO WARRANTY !!
 ***/


/// !!! Include your device specific header file here !!! ///
#include "DMX-lib_mega644p_usart1.io.h"
/// !!! Include your device specific header file here !!! ///

#ifndef F_CPU
	#error "This DMX libary does need to have F_CPU defined. Please define F_CPU with the MCU clock frequency in Hz! Aborting."
#endif

#ifndef DMX_UBRR_VAL
#if (F_CPU == 8000000UL)
#define DMX_UBRR_VAL	1 
// (250kbps at F_CPU 8e6, see datasheet ATmegaXX4P p.198
#warning "DMX_UBRR_VAL not defined. Assumed '1' (250kbps at 8MHz). DMX communication might not work."
#elif (F_CPU == 16000000UL)
#define DMX_UBRR_VAL	3 
// (250kbps at F_CPU 16e6, see datasheet ATmegaXX4P p.198
#warning "DMX_UBRR_VAL not defined. Assumed '3' (250kbps at 16MHz). DMX communication might not work."
#else
#error "DMX_UBRR_VAL is not defined. Aborting."
#endif
#endif



/* 
 * Receiver and Transmitter settings
 */


// Receiver width (set these constant to zero if you do not want to use the receiver)
#ifndef CHANNEL_TO_RX
	#define CHANNEL_TO_RX		2
#endif

// Transmitter width (set these constant to zero if you do not want to use the transmitter)
#ifndef CHANNEL_TO_TX
	#define CHANNEL_TO_TX		128
#endif

// Receiver incomming delay (in ms)
#define RX_INCOMMING_DELAY_MS 	2000

// Transmitter BREAK length (in us)
#define TX_BREAK_LENGTH_US		110
// Transmitter MAB length (in us)
#define TX_MAB_LENGTH_US		20

// Starbyte Codes (default: 0x00)
#define RX_STARTBYTE_VALUE		0x00
#define TX_STARTBYTE_VALUE		0x00







/* 
 * Constant calculations: Timing values (do not change!!)
 */
// Info: using 16bit timer with prescaler 8

// Calculate incomming delay overflows
#define TEMP_RX_INCOMMING_DELAY		(((F_CPU/8)/65536) * (RX_INCOMMING_DELAY_MS * (1/1000)))
#if (TEMP_RX_INCOMMING_DELAY > 255)
	#define RX_INCOMMING_DELAY		255
	#warning "RX_INCOMMING_DELAY is greater than 255 and therefor not fitting into an uint8_t. You may choose a lower value for RX_INCOMMING_DELAY_MS!"
#else
	#define RX_INCOMMING_DELAY		TEMP_RX_INCOMMING_DELAY
#endif

// Calculate BREAK length timing
#define TEMP_TX_BREAK_LENGTH		((F_CPU/8) * (TX_BREAK_LENGTH_US * (1/100000)))
#if (TEMP_TX_BREAK_LENGTH > 65000)
	#error "TX_BREAK_LENGTH is greater than 65000 and therefor to high. Check your value for TX_BREAK_LENGTH_US; it should be about 100us. Aborting."
#else
	#define TX_BREAK_LENGTH			TEMP_TX_BREAK_LENGTH
#endif

// Calculate MAB length timing
#define TEMP_TX_MAB_LENGTH			((F_CPU/8) * (TX_MAB_LENGTH_US * (1/100000)))
#if (TEMP_TX_MAB_LENGTH > 8000)
	#error "TX_MAB_LENGTH is greater than 8000 and therefor to high. Check your value for TX_MAB_LENGTH_US; it should be about 8us to 12us. Aborting."
#else
	#define TX_MAB_LENGTH			TEMP_TX_MAB_LENGTH
#endif



/* 
 * Type definitions
 */

#if (CHANNEL_TO_RX > 0)
// Receive variable definitions
enum tDmxRxStates {RX_NONE=0, RX_RESET, RX_STARTBYTE, RX_DATABYTESB, RX_DATA, RX_DATABYTESA, RX_FAIL}; // RX_RESET = (BREAK + MAB)
volatile struct tDmxRx {
	
	enum tDmxRxStates state;
	uint8_t comming;
	uint16_t startAddr;
	uint16_t channelCounter;
	
	uint8_t values[CHANNEL_TO_RX];
	
} DmxRx;
#endif

#if (CHANNEL_TO_TX > 0)
// Transceive variable definitions
enum tDmxTxStates {TX_NONE=0, TX_BREAK, TX_MAB, TX_STARTBYTE, TX_DATA};
volatile struct tDmxTx {
	
	enum tDmxTxStates state;
	uint16_t channelCounter;
	
	uint8_t values[CHANNEL_TO_TX];
	
} DmxTx;
#endif


void dmx_init(void){
	
	DMX_UBRRH = ((DMX_UBRR_VAL & 0x0F00)>>8);
	DMX_UBRRL = DMX_UBRR_VAL & 0x00FF;
	
	DMX_UCSRC = (1<<DMX_USBS) | (1<<DMX_UCSZ1) | (1<<DMX_UCSZ0); // 8n2
	
	// enable timer1 for: (RX: data incomming check) || (TX: BREAK and MAB timings)
	DMX_TCCR1B |= (1<<DMX_CS11); // run with f_cpu/8
	
	
	#if (CHANNEL_TO_RX > 0)
		#warning Using DMX RX
	DMX_RX_DDR &= ~(1<<DMX_RX_PIN);
	DMX_UCSRB |= (1<<DMX_RXEN) | (1<<DMX_RXCIE);
	DMX_TIMSK1 |= (1<<DMX_TOIE1); // enable overflow interrupt
	DmxRx.state = RX_NONE;
	#endif
	
	#if (CHANNEL_TO_TX > 0)
		#warning Using DMX TX
	DMX_TX_DDR |= (1<<DMX_TX_PIN);
	DMX_TX_PORT |= (1<<DMX_TX_PIN); // Idle status is high
	DMX_UCSRB |= (1<<DMX_TXEN);
	DMX_UCSRB |= (1<<DMX_UDRIE);
	  // -> TIMER interrupts are temporally enabled in UDRE interrupt routines for BREAK and MAB timings
	DmxTx.state = TX_NONE;
	#endif
	
}



/* 
 * Receive (RX Routines)
 */
#if (CHANNEL_TO_RX > 0)

// receiver incomming check routine
ISR(DMX_TIMER1_OVF_vect){
	
	if(DmxRx.comming <= 1){ // stil waiting for something comming in - not ok, so no dmx incomming
		DmxRx.comming = 0;
		DmxRx.state = RX_NONE;
	}
	else{
		DmxRx.comming--; // move in non-comming direction
	}
	
}

// receive routine
ISR(DMX_USART_RX_vect){
	DmxRx.comming = RX_INCOMMING_DELAY; // here is something comming in, reset to highest value
	
	uint8_t rxstatus = DMX_UCSRA; // read status
	uint8_t rxval = DMX_UDR; // read data and empty buffer
	
	
	if(rxstatus & (1<<DMX_FE)){
		
		if(DmxRx.state == RX_RESET || // more than one FE is detected as failure
		  DmxRx.state == RX_DATABYTESB){ // there were no receive of dmx_data. Startaddress is to high/not enough channels on the line
		   DmxRx.state = RX_FAIL;
		}
		else{
			   DmxRx.state = RX_RESET;
		}
		return;
	}
	
	if(rxstatus & (1<<DMX_DOR)){ // dataoverrun is detected as fail
		DmxRx.state = RX_FAIL;
		return;
	}
	
	
	
	switch(DmxRx.state){
		case RX_RESET:
			
			DmxRx.channelCounter = 0;
			
			if(rxval == RX_STARTBYTE_VALUE){ // startbyte with RX_STARTBYTE_VALUE is detected as ok. otherwise previous FE was not reset but failure
				if(DmxRx.startAddr == 1){ // do not wait next time, catch next (first) incomming value.
					DmxRx.state = RX_DATA;
				}
				else{
					DmxRx.state = RX_STARTBYTE;
				}
			}
			else{
				DmxRx.state = RX_FAIL;
			}
			
		break;
		case RX_STARTBYTE:
		case RX_DATABYTESB:
			
			DmxRx.channelCounter++;
			
			if(DmxRx.channelCounter == (DmxRx.startAddr - 1)){ // receive next channel
				DmxRx.state = RX_DATA;
			}
			
		break;
		case RX_DATA:
			
			DmxRx.channelCounter++;
			
			uint8_t ch_to_rx = (uint8_t) DmxRx.channelCounter - DmxRx.startAddr;
			
			DmxRx.values[ch_to_rx] = rxval;
			
			if(ch_to_rx >= CHANNEL_TO_RX){
				DmxRx.state = RX_DATABYTESA; // channels received, just wait now.
			}
			
		break;
		default: break;
	}
}

#endif


/*
 *  Send (TX Routines)
 */
#if (CHANNEL_TO_TX > 0)


inline void _setOCRval(uint16_t addToNow){	// NOTICE: use only with disabled interrupts to have atomic 16bit operations!
	uint16_t temp;
	temp = DMX_TCNT1L; // read counter
	temp |= (DMX_TCNT1H<<8);
	temp = temp + addToNow; // add wait time to counter value
	DMX_OCR1AH = ((temp & 0xFF00)>>8); // write new value for compare match
	DMX_OCR1AL = temp & 0x00FF;
}


// BREAK and MAB generation routine
ISR(DMX_TIMER1_COMPA_vect){
	PORTD &= ~(1<<PD5);
	switch(DmxTx.state){
		case TX_BREAK:
			
			DMX_TX_PORT |= (1<<DMX_TX_PIN); // start MAB
			_setOCRval(TX_MAB_LENGTH);
			
			DmxTx.state = TX_MAB;
			
		break;
		case TX_MAB:
						
			DMX_TIMSK1 &= ~(1<<DMX_OCIE1A); // disable compare match interrupt
			
			DMX_TX_PORT &= ~(1<<DMX_TX_PIN); // End MAB, start start-bit of startbyte
			DMX_UCSRB |= (1<<DMX_TXEN); // give UART pin-control
			DMX_UCSRB |= (1<<DMX_UDRIE); // enable UART transmitter and empty-interrupt
			
			DmxTx.state = TX_STARTBYTE;
			
		break;
		default: break;
	}
}


// Dmx Data Tranceive Routine
ISR(DMX_USART_UDRE_vect){
	switch(DmxTx.state){
		case TX_NONE:
		case TX_DATA:
			if(DmxTx.channelCounter == CHANNEL_TO_TX || DmxTx.state == TX_NONE){
				
				DMX_UCSRB &= ~((1<<DMX_TXEN) | (1<<DMX_UDRIE)); // disable transmitter and UDRE interrupts
				DMX_TX_PORT &= ~(1<<DMX_TX_PIN); // set to low an therefor start BREAK
				
				DmxTx.state = TX_BREAK;
				
				_setOCRval(TX_BREAK_LENGTH);
				DMX_TIMSK1 |= (1<<DMX_OCIE1A); // enable compare match interrupt
				
				return;
			}
			
			DMX_UDR = DmxTx.values[DmxTx.channelCounter];
			DmxTx.channelCounter++;
			
		break;
		case TX_STARTBYTE:
			
			DMX_UDR = TX_STARTBYTE_VALUE;
			DmxTx.state = TX_DATA;
			DmxTx.channelCounter = 0;
			
		break;
		default: break;
	}
}

#endif
